import asyncio
import base64
import hashlib
import hmac
import json
from typing import Any, Awaitable, Callable, Dict, Optional, Text

import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    HTTPResponse,
    Request,
    Websocket,
    response,
)

from rasa.core.channels import UserMessage
from rasa.core.channels.voice_ready.utils import CallParameters
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import (
    call_state,
)
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)

"""
Genesys throws a rate limit error with too many audio messages.
To avoid this, we buffer the audio messages and send them in chunks.

- global.inbound.binary.average.rate.per.second: 5
The allowed average rate per second of inbound binary data

- global.inbound.binary.max: 25
The maximum number of inbound binary data messages
that can be sent instantaneously

https://developer.genesys.cloud/organization/organization/limits#audiohook

The maximum binary message size is not mentioned
in the documentation but observed in their example app
https://github.com/GenesysCloudBlueprints/audioconnector-server-reference-implementation
"""
MAXIMUM_BINARY_MESSAGE_SIZE = 64000  # 64KB
HEADER_API_KEY = "X-Api-Key"
logger = structlog.get_logger(__name__)


def map_call_params(data: Dict[Text, Any]) -> CallParameters:
    """Map the twilio stream parameters to the CallParameters dataclass."""
    parameters = data["parameters"]
    participant = parameters["participant"]
    # sent as {"ani": "tel:+491604697810"}
    ani = participant.get("ani", "")
    user_phone = ani.split(":")[-1] if ani else ""

    return CallParameters(
        call_id=parameters.get("conversationId", ""),
        user_phone=user_phone,
        bot_phone=participant.get("dnis", ""),
    )


class GenesysOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "genesys"

    async def send_audio_bytes(
        self, recipient_id: str, audio_bytes: RasaAudioBytes
    ) -> None:
        await self.voice_websocket.send(audio_bytes)

    async def send_marker_message(self, recipient_id: str) -> None:
        """
        Send a message that marks positions in the audio stream.
        Genesys does not support this feature, so we do nothing here.
        """
        pass


class GenesysInputChannel(VoiceInputChannel):
    @classmethod
    def name(cls) -> str:
        return "genesys"

    def __init__(
        self, api_key: Text, client_secret: Optional[Text], *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.api_key = api_key
        self.client_secret = client_secret

    @classmethod
    def from_credentials(
        cls, credentials: Optional[Dict[str, Any]]
    ) -> VoiceInputChannel:
        if not credentials:
            raise ValueError("No credentials given for Genesys voice channel.")

        if not credentials.get("api_key"):
            raise ValueError("No API key given for Genesys voice channel (api_key).")

        return cls(
            api_key=credentials["api_key"],
            client_secret=credentials.get("client_secret"),
            server_url=credentials["server_url"],
            asr_config=credentials["asr"],
            tts_config=credentials["tts"],
            monitor_silence=credentials.get("monitor_silence", False),
        )

    def _ensure_channel_data_initialized(self) -> None:
        """Initialize Genesys-specific channel data if not already present.

        Genesys requires the server and client each maintain a
        monotonically increasing message sequence number.
        """
        if "server_sequence_number" not in call_state.channel_data:
            call_state.channel_data["server_sequence_number"] = 0
        if "client_sequence_number" not in call_state.channel_data:
            call_state.channel_data["client_sequence_number"] = 0

    def _get_next_sequence(self) -> int:
        """
        Get the next message sequence number
        Rasa == Server
        Genesys == Client

        Genesys requires the server and client each maintain a
        monotonically increasing message sequence number.
        """
        self._ensure_channel_data_initialized()
        call_state.channel_data["server_sequence_number"] += 1
        return call_state.channel_data["server_sequence_number"]

    def _get_last_client_sequence(self) -> int:
        """Get the last client(Genesys) sequence number."""
        self._ensure_channel_data_initialized()
        return call_state.channel_data["client_sequence_number"]

    def _update_client_sequence(self, seq: int) -> None:
        """Update the client(Genesys) sequence number."""
        self._ensure_channel_data_initialized()

        if seq - call_state.channel_data["client_sequence_number"] != 1:
            logger.warning(
                "genesys.update_client_sequence.sequence_gap",
                received_seq=seq,
                last_seq=call_state.channel_data["client_sequence_number"],
            )
        call_state.channel_data["client_sequence_number"] = seq

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        return RasaAudioBytes(input_bytes)

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        """Call Parameters are collected during the open event."""
        async for message in channel_websocket:
            data = json.loads(message)
            self._update_client_sequence(data["seq"])
            if data.get("type") == "open":
                call_params = await self.handle_open(channel_websocket, data)
                return call_params
            else:
                logger.error("genesys.receive.unexpected_initial_message", message=data)

        return None

    def map_input_message(
        self,
        message: Any,
        ws: Websocket,
    ) -> VoiceChannelAction:
        # if message is binary, it's audio
        if isinstance(message, bytes):
            return NewAudioAction(self.channel_bytes_to_rasa_audio_bytes(message))
        else:
            # process text message
            data = json.loads(message)
            self._update_client_sequence(data["seq"])
            msg_type = data.get("type")
            if msg_type == "close":
                logger.info("genesys.handle_close", message=data)
                self.handle_close(ws, data)
                return EndConversationAction()
            elif msg_type == "ping":
                logger.info("genesys.handle_ping", message=data)
                self.handle_ping(ws, data)
            elif msg_type == "playback_started":
                logger.debug("genesys.handle_playback_started", message=data)
                call_state.is_bot_speaking = True  # type: ignore[attr-defined]
            elif msg_type == "playback_completed":
                logger.debug("genesys.handle_playback_completed", message=data)
                call_state.is_bot_speaking = False  # type: ignore[attr-defined]
                if call_state.should_hangup:
                    logger.info("genesys.hangup")
                    self.disconnect(ws, data)
                    # the conversation should continue until
                    # we receive a close message from Genesys
            elif msg_type == "dtmf":
                logger.info("genesys.handle_dtmf", message=data)
            elif msg_type == "error":
                logger.warning("genesys.handle_error", message=data)
            else:
                logger.warning("genesys.map_input_message.unknown_type", message=data)

        return ContinueConversationAction()

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        return GenesysOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
            min_buffer_size=MAXIMUM_BINARY_MESSAGE_SIZE // 2,
        )

    async def handle_open(self, ws: Websocket, message: dict) -> CallParameters:
        """Handle initial open transaction from Genesys."""
        call_parameters = map_call_params(message)
        params = message["parameters"]
        media_options = params.get("media", [])

        # Send opened response
        if media_options:
            logger.info("genesys.handle_open", media_parameter=media_options[0])
            response = {
                "version": "2",
                "type": "opened",
                "seq": self._get_next_sequence(),
                "clientseq": self._get_last_client_sequence(),
                "id": message.get("id"),
                "parameters": {"startPaused": False, "media": [media_options[0]]},
            }
            logger.debug("genesys.handle_open.opened", response=response)
            await ws.send(json.dumps(response))
        else:
            logger.warning(
                "genesys.handle_open.no_media_formats", client_message=message
            )
        return call_parameters

    def handle_ping(self, ws: Websocket, message: dict) -> None:
        """Handle ping message from Genesys."""
        response = {
            "version": "2",
            "type": "pong",
            "seq": self._get_next_sequence(),
            "clientseq": message.get("seq"),
            "id": message.get("id"),
            "parameters": {},
        }
        logger.debug("genesys.handle_ping.pong", response=response)
        _schedule_ws_task(ws.send(json.dumps(response)))

    def handle_close(self, ws: Websocket, message: dict) -> None:
        """Handle close message from Genesys."""
        response = {
            "version": "2",
            "type": "closed",
            "seq": self._get_next_sequence(),
            "clientseq": self._get_last_client_sequence(),
            "id": message.get("id"),
            "parameters": message.get("parameters", {}),
        }
        logger.debug("genesys.handle_close.closed", response=response)

        _schedule_ws_task(ws.send(json.dumps(response)))

    def disconnect(self, ws: Websocket, data: dict) -> None:
        """
        Send disconnect message to Genesys.

        https://developer.genesys.cloud/devapps/audiohook/protocol-reference#disconnect
        It should be used to hangup the call.
        Genesys will respond with a "close" message to us
        that is handled by the handle_close method.
        """
        message = {
            "version": "2",
            "type": "disconnect",
            "seq": self._get_next_sequence(),
            "clientseq": self._get_last_client_sequence(),
            "id": data.get("id"),
            "parameters": {
                "reason": "completed",
                # arbitrary values can be sent here
            },
        }
        logger.debug("genesys.disconnect", message=message)
        _schedule_ws_task(ws.send(json.dumps(message)))

    def _calculate_signature(self, request: Request) -> str:
        """Calculate the signature using request data."""
        org_id = request.headers.get("Audiohook-Organization-Id")
        session_id = request.headers.get("Audiohook-Session-Id")
        correlation_id = request.headers.get("Audiohook-Correlation-Id")
        api_key = request.headers.get(HEADER_API_KEY)

        # order of components is important!
        components = [
            ("@request-target", "/webhooks/genesys/websocket"),
            ("audiohook-session-id", session_id),
            ("audiohook-organization-id", org_id),
            ("audiohook-correlation-id", correlation_id),
            (HEADER_API_KEY.lower(), api_key),
            ("@authority", self.server_url),
        ]

        # Create signature base string
        signing_string = ""
        for name, value in components:
            signing_string += f'"{name}": {value}\n'

        # Add @signature-params
        signature_input = request.headers["Signature-Input"]
        _, params_str = signature_input.split("=", 1)
        signing_string += f'"@signature-params": {params_str}'

        # Calculate the HMAC signature
        key_bytes = base64.b64decode(self.client_secret)
        signature = hmac.new(
            key_bytes, signing_string.encode("utf-8"), hashlib.sha256
        ).digest()
        return base64.b64encode(signature).decode("utf-8")

    async def _verify_signature(self, request: Request) -> bool:
        """Verify the HTTP message signature from Genesys."""
        if not self.client_secret:
            logger.info(
                "genesys.verify_signature.no_client_secret",
                event_info="Signature verification skipped",
            )
            return True  # Skip verification if no client secret

        signature = request.headers.get("Signature")
        signature_input = request.headers.get("Signature-Input")
        if not signature or not signature_input:
            logger.error("genesys.signature.missing_signature_header")
            return False

        try:
            actual_signature = signature.split("=", 1)[1].strip(':"')
            expected_signature = self._calculate_signature(request)
            return hmac.compare_digest(
                expected_signature.encode("utf-8"), actual_signature.encode("utf-8")
            )
        except Exception as e:
            logger.exception("genesys.signature.verification_error", error=e)
            return False

    def _ensure_required_headers(self, request: Request) -> bool:
        """Ensure required headers are present in the request."""
        required_headers = [
            "Audiohook-Organization-Id",
            "Audiohook-Correlation-Id",
            "Audiohook-Session-Id",
            HEADER_API_KEY,
        ]

        missing_headers = [
            header for header in required_headers if header not in request.headers
        ]

        if missing_headers:
            logger.error(
                "genesys.missing_required_headers",
                missing_headers=missing_headers,
            )
            return False
        return True

    def _ensure_api_key(self, request: Request) -> bool:
        """Ensure the API key is present in the request."""
        api_key = request.headers.get(HEADER_API_KEY)
        if not hmac.compare_digest(str(self.api_key), str(api_key)):
            return False
        return True

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        """Defines a Sanic blueprint for the voice input channel."""
        blueprint = Blueprint("genesys", __name__)

        @blueprint.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @blueprint.websocket("/websocket")  # type: ignore[misc]
        async def receive(request: Request, ws: Websocket) -> None:
            logger.debug(
                "genesys.receive",
                audiohook_session_id=request.headers.get("audiohook-session-id"),
            )

            # verify signature
            if not await self._verify_signature(request):
                logger.error("genesys.receive.invalid_signature")
                await ws.close(code=1008, reason="Invalid signature")
                return

            # ensure required headers are present
            if not self._ensure_required_headers(request):
                await ws.close(code=1002, reason="Missing required headers")
                return

            # ensure API key is correct
            if not self._ensure_api_key(request):
                logger.error(
                    "genesys.receive.invalid_api_key",
                    invalid_api_key=request.headers.get(HEADER_API_KEY),
                )
                await ws.close(code=1008, reason="Invalid API key")
                return

            # process audio streaming
            logger.info("genesys.receive", message="Starting audio streaming")
            try:
                await self.run_audio_streaming(on_new_message, ws)
            except Exception as e:
                logger.exception(
                    "genesys.receive",
                    message="Error during audio streaming",
                    error=e,
                )
                await ws.close(code=1011, reason="Error during audio streaming")
                raise

        return blueprint


def _schedule_ws_task(coro: Awaitable[Any]) -> None:
    """Helper function to schedule a coroutine in the event loop.

    Args:
        coro: The coroutine to schedule
    """
    loop = asyncio.get_running_loop()
    loop.call_soon_threadsafe(lambda: loop.create_task(coro))



from contextlib import contextmanager
from datetime import datetime
from typing import Callable, Generator, List, Union

import structlog

from rasa.dialogue_understanding.commands.prompt_command import PromptCommand
from rasa.dialogue_understanding.generator import LLMBasedCommandGenerator
from rasa.shared.providers.llm.llm_response import LLMResponse

structlogger = structlog.get_logger()


def commands_as_string(commands: List[PromptCommand], delimiter: str = "\n") -> str:
    return delimiter.join([command.to_dsl() for command in commands])


def make_mock_invoke_llm(commands: str) -> Callable:
    """Capture the `commands` in a closure so the resulting async function
    can use it as its response.

    Args:
        commands: The commands to return from the mock LLM call.
    """

    async def _mock_invoke_llm(
        self: LLMBasedCommandGenerator, prompt: Union[List[dict], List[str], str]
    ) -> LLMResponse:
        structlogger.debug(
            f"LLM call intercepted, response mocked. "
            f"Responding with the following commands: '{commands}' "
            f"to the prompt: {prompt}"
        )
        fake_response_dict = {
            "id": "",
            "choices": [commands],
            "created": int(datetime.now().timestamp()),
            "model": "mocked-llm",
        }
        return LLMResponse.from_dict(fake_response_dict)

    return _mock_invoke_llm


@contextmanager
def patch_invoke_llm_in_generators(mock_impl: Callable) -> Generator:
    """Replace CommandGenerator.invoke_llm in the base class AND in all
    current subclasses (recursively).  Everything is restored on exit.
    """
    originals = {}

    def collect(cls: type[LLMBasedCommandGenerator]) -> None:
        # store current attribute, then recurse
        originals[cls] = cls.invoke_llm
        for sub in cls.__subclasses__():
            collect(sub)

    # collect every existing subclass of CommandGenerator
    collect(LLMBasedCommandGenerator)  # type: ignore[type-abstract]

    try:
        # apply the monkey-patch everywhere
        for cls in originals:
            cls.invoke_llm = mock_impl  # type: ignore[assignment]
        yield
    finally:
        # restore originals (even if an exception happened)
        for cls, orig in originals.items():
            cls.invoke_llm = orig  # type: ignore[assignment]



from __future__ import annotations

import asyncio
import json
import logging
import os
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Dict, Optional, Text, Union

import rasa.shared.utils.common
from rasa.core.constants import DEFAULT_LOCK_LIFETIME
from rasa.core.lock import TicketLock
from rasa.shared.exceptions import ConnectionException, RasaException
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


def _get_lock_lifetime() -> int:
    return int(os.environ.get("TICKET_LOCK_LIFETIME", 0)) or DEFAULT_LOCK_LIFETIME


LOCK_LIFETIME = _get_lock_lifetime()
DEFAULT_SOCKET_TIMEOUT_IN_SECONDS = 10

DEFAULT_REDIS_LOCK_STORE_KEY_PREFIX = "lock:"


# noinspection PyUnresolvedReferences
class LockError(RasaException):
    """Exception that is raised when a lock cannot be acquired.

    Attributes:
         message (str): explanation of which `conversation_id` raised the error
    """

    pass


class LockStore:
    """Base class for ticket locks."""

    @staticmethod
    def create(obj: Union[LockStore, EndpointConfig, None]) -> LockStore:
        """Factory to create a lock store."""
        if isinstance(obj, LockStore):
            return obj

        try:
            return _create_from_endpoint_config(obj)
        except ConnectionError as error:
            raise ConnectionException("Cannot connect to lock store.") from error

    @staticmethod
    def create_lock(conversation_id: Text) -> TicketLock:
        """Create a new `TicketLock` for `conversation_id`."""
        return TicketLock(conversation_id)

    def get_lock(self, conversation_id: Text) -> Optional[TicketLock]:
        """Fetch lock for `conversation_id` from storage."""
        raise NotImplementedError

    def delete_lock(self, conversation_id: Text) -> None:
        """Delete lock for `conversation_id` from storage."""
        raise NotImplementedError

    def save_lock(self, lock: TicketLock) -> None:
        """Commit `lock` to storage."""
        raise NotImplementedError

    def issue_ticket(
        self, conversation_id: Text, lock_lifetime: float = LOCK_LIFETIME
    ) -> int:
        """Issue new ticket with `lock_lifetime` for lock associated with
        `conversation_id`.

        Creates a new lock if none is found.
        """
        logger.debug(f"Issuing ticket for conversation '{conversation_id}'.")
        try:
            lock = self.get_or_create_lock(conversation_id)
            ticket = lock.issue_ticket(lock_lifetime)
            self.save_lock(lock)

            return ticket
        except Exception as e:
            raise LockError(f"Error while acquiring lock. Error:\n{e}")

    @asynccontextmanager
    async def lock(
        self,
        conversation_id: Text,
        lock_lifetime: float = LOCK_LIFETIME,
        wait_time_in_seconds: float = 1,
    ) -> AsyncGenerator[TicketLock, None]:
        """Acquire lock with lifetime `lock_lifetime`for `conversation_id`.

        Try acquiring lock with a wait time of `wait_time_in_seconds` seconds
        between attempts. Raise a `LockError` if lock has expired.
        """
        ticket = self.issue_ticket(conversation_id, lock_lifetime)
        try:
            yield await self._acquire_lock(
                conversation_id, ticket, wait_time_in_seconds
            )
        finally:
            self.cleanup(conversation_id, ticket)

    async def _acquire_lock(
        self, conversation_id: Text, ticket: int, wait_time_in_seconds: float
    ) -> TicketLock:
        logger.debug(f"Acquiring lock for conversation '{conversation_id}'.")
        while True:
            # fetch lock in every iteration because lock might no longer exist
            lock = self.get_lock(conversation_id)

            # exit loop if lock does not exist anymore (expired)
            if not lock:
                break

            # acquire lock if it isn't locked
            if not lock.is_locked(ticket):
                logger.debug(f"Acquired lock for conversation '{conversation_id}'.")
                return lock

            items_before_this = ticket - (lock.now_serving or 0)

            logger.debug(
                f"Failed to acquire lock for conversation ID '{conversation_id}' "
                f"because {items_before_this} other item(s) for this "
                f"conversation ID have to be finished processing first. "
                f"Retrying in {wait_time_in_seconds} seconds ..."
            )

            # sleep and update lock
            await asyncio.sleep(wait_time_in_seconds)
            self.update_lock(conversation_id)

        raise LockError(
            f"Could not acquire lock for conversation_id '{conversation_id}'."
        )

    def update_lock(self, conversation_id: Text) -> None:
        """Fetch lock for `conversation_id`, remove expired tickets and save lock."""
        lock = self.get_lock(conversation_id)
        if lock:
            lock.remove_expired_tickets()
            self.save_lock(lock)

    def get_or_create_lock(self, conversation_id: Text) -> TicketLock:
        """Fetch existing lock for `conversation_id`.

        Alternatively, create a new one if it doesn't exist.
        """
        existing_lock = self.get_lock(conversation_id)

        if existing_lock:
            return existing_lock

        return self.create_lock(conversation_id)

    def is_someone_waiting(self, conversation_id: Text) -> bool:
        """Return whether someone is waiting for lock for this `conversation_id`."""
        lock = self.get_lock(conversation_id)
        if lock:
            return lock.is_someone_waiting()

        return False

    def finish_serving(self, conversation_id: Text, ticket_number: int) -> None:
        """Finish serving ticket with `ticket_number` for `conversation_id`.

        Removes ticket from lock and saves lock.
        """
        lock = self.get_lock(conversation_id)
        if lock:
            lock.remove_ticket_for(ticket_number)
            self.save_lock(lock)

    def cleanup(self, conversation_id: Text, ticket_number: int) -> None:
        """Remove lock for `conversation_id` if no one is waiting."""
        self.finish_serving(conversation_id, ticket_number)
        if not self.is_someone_waiting(conversation_id):
            self.delete_lock(conversation_id)

    @staticmethod
    def _log_deletion(conversation_id: Text, deletion_successful: bool) -> None:
        if deletion_successful:
            logger.debug(f"Deleted lock for conversation '{conversation_id}'.")
        else:
            logger.debug(f"Could not delete lock for conversation '{conversation_id}'.")


class RedisLockStore(LockStore):
    """Redis store for ticket locks."""

    def __init__(
        self,
        host: Text = "localhost",
        port: int = 6379,
        db: int = 1,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        use_ssl: bool = False,
        ssl_certfile: Optional[Text] = None,
        ssl_keyfile: Optional[Text] = None,
        ssl_ca_certs: Optional[Text] = None,
        key_prefix: Optional[Text] = None,
        socket_timeout: float = DEFAULT_SOCKET_TIMEOUT_IN_SECONDS,
    ) -> None:
        """Create a lock store which uses Redis for persistence.

        Args:
            host: The host of the redis server.
            port: The port of the redis server.
            db: The name of the database within Redis which should be used by Rasa
                Open Source.
            username: The username which should be used for authentication with the
                Redis database.
            password: The password which should be used for authentication with the
                Redis database.
            use_ssl: `True` if SSL should be used for the connection to Redis.
            ssl_certfile: Path to the SSL certificate file.
            ssl_keyfile: Path to the SSL private key file.
            ssl_ca_certs: Path to the SSL CA certificate file.
            key_prefix: prefix to prepend to all keys used by the lock store. Must be
                alphanumeric.
            socket_timeout: Timeout in seconds after which an exception will be raised
                in case Redis doesn't respond within `socket_timeout` seconds.
        """
        import redis

        self.red = redis.StrictRedis(
            host=host,
            port=int(port),
            db=int(db),
            username=username,
            password=password,
            ssl=use_ssl,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            ssl_ca_certs=ssl_ca_certs,
            socket_timeout=socket_timeout,
        )

        self.key_prefix = DEFAULT_REDIS_LOCK_STORE_KEY_PREFIX
        if key_prefix:
            logger.debug(f"Setting non-default redis key prefix: '{key_prefix}'.")
            self._set_key_prefix(key_prefix)

        super().__init__()

    def _set_key_prefix(self, key_prefix: Text) -> None:
        if isinstance(key_prefix, str) and key_prefix.isalnum():
            self.key_prefix = key_prefix + ":" + DEFAULT_REDIS_LOCK_STORE_KEY_PREFIX
        else:
            logger.warning(
                f"Omitting provided non-alphanumeric redis key prefix: '{key_prefix}'. "
                f"Using default '{self.key_prefix}' instead."
            )

    def get_lock(self, conversation_id: Text) -> Optional[TicketLock]:
        """Retrieves lock (see parent docstring for more information)."""
        serialised_lock = self.red.get(self.key_prefix + conversation_id)
        if serialised_lock:
            return TicketLock.from_dict(json.loads(serialised_lock))

        return None

    def delete_lock(self, conversation_id: Text) -> None:
        """Deletes lock for conversation ID."""
        deletion_successful = self.red.delete(self.key_prefix + conversation_id)
        self._log_deletion(conversation_id, deletion_successful)

    def save_lock(self, lock: TicketLock) -> None:
        self.red.set(self.key_prefix + lock.conversation_id, lock.dumps())


class InMemoryLockStore(LockStore):
    """In-memory store for ticket locks."""

    def __init__(self) -> None:
        """Initialise dictionary of locks."""
        self.conversation_locks: Dict[Text, TicketLock] = {}
        super().__init__()

    def get_lock(self, conversation_id: Text) -> Optional[TicketLock]:
        """Get lock for conversation if it exists."""
        return self.conversation_locks.get(conversation_id)

    def delete_lock(self, conversation_id: Text) -> None:
        """Delete lock for conversation."""
        deleted_lock = self.conversation_locks.pop(conversation_id, None)
        self._log_deletion(
            conversation_id, deletion_successful=deleted_lock is not None
        )

    def save_lock(self, lock: TicketLock) -> None:
        """Save lock in store."""
        self.conversation_locks[lock.conversation_id] = lock


def _create_from_endpoint_config(
    endpoint_config: Optional[EndpointConfig] = None,
) -> LockStore:
    """Given an endpoint configuration, create a proper `LockStore` object."""
    if (
        endpoint_config is None
        or endpoint_config.type is None
        or endpoint_config.type == "in_memory"
    ):
        # this is the default type if no lock store type is set

        lock_store: LockStore = InMemoryLockStore()
    elif endpoint_config.type == "redis":
        lock_store = RedisLockStore(host=endpoint_config.url, **endpoint_config.kwargs)
    elif endpoint_config.type == "concurrent_redis":
        from rasa.core.concurrent_lock_store import ConcurrentRedisLockStore

        lock_store = ConcurrentRedisLockStore(endpoint_config=endpoint_config)
    else:
        lock_store = _load_from_module_name_in_endpoint_config(endpoint_config)

    logger.debug(f"Connected to lock store '{lock_store.__class__.__name__}'.")

    return lock_store


def _load_from_module_name_in_endpoint_config(
    endpoint_config: EndpointConfig,
) -> LockStore:
    """Retrieve a `LockStore` based on its class name."""
    try:
        lock_store_class = rasa.shared.utils.common.class_from_module_path(
            endpoint_config.type
        )
        return lock_store_class(endpoint_config=endpoint_config)
    except (AttributeError, ImportError) as e:
        raise Exception(
            f"Could not find a class based on the module path "
            f"'{endpoint_config.type}'. Failed to create a `LockStore` "
            f"instance. Error: {e}"
        )



import copy
from typing import Any, Dict, List, Literal, Optional, Text

import structlog

import rasa.shared.utils.io
from rasa.dialogue_understanding.commands import (
    CannotHandleCommand,
    Command,
    ErrorCommand,
    SetSlotCommand,
)
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.generator import LLMBasedCommandGenerator
from rasa.dialogue_understanding.generator.command_parser import (
    parse_commands as parse_commands_using_command_parsers,
)
from rasa.dialogue_understanding.generator.constants import (
    COMMAND_PROMPT_FILE_NAME,
    DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    FLOW_RETRIEVAL_KEY,
    LLM_BASED_COMMAND_GENERATOR_CONFIG_FILE,
    LLM_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CLAUDE_3_5_SONNET_20240620,
    MODEL_NAME_GPT_4O_2024_11_20,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
    USER_INPUT_CONFIG_KEY,
)
from rasa.dialogue_understanding.generator.flow_retrieval import FlowRetrieval
from rasa.dialogue_understanding.stack.utils import top_flow_frame
from rasa.dialogue_understanding.utils import (
    add_commands_to_message_parse_data,
    add_prompt_to_message_parse_data,
)
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    ANTHROPIC_PROVIDER,
    AWS_BEDROCK_PROVIDER,
    AZURE_OPENAI_PROVIDER,
    EMBEDDINGS_CONFIG_KEY,
    MAX_COMPLETION_TOKENS_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    ROUTE_TO_CALM_SLOT,
    TEMPERATURE_CONFIG_KEY,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.nlu.constants import LLM_COMMANDS, LLM_PROMPT, TEXT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.providers.llm.llm_response import LLMResponse
from rasa.shared.utils.constants import (
    LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON,
    LOG_COMPONENT_SOURCE_METHOD_INIT,
)
from rasa.shared.utils.io import deep_container_fingerprint
from rasa.shared.utils.llm import (
    allowed_values_for_slot,
    get_default_prompt_template_based_on_model,
    get_prompt_template,
    resolve_model_client_config,
    sanitize_message_for_prompt,
    tracker_as_readable_transcript,
)
from rasa.utils.beta import BetaNotEnabledException, ensure_beta_feature_is_enabled
from rasa.utils.log_utils import log_llm

structlogger = structlog.get_logger()


DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: MODEL_NAME_GPT_4O_2024_11_20,
    TEMPERATURE_CONFIG_KEY: 0.0,
    MAX_COMPLETION_TOKENS_CONFIG_KEY: DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    TIMEOUT_CONFIG_KEY: 7,
}

MODEL_PROMPT_MAPPER = {
    f"{OPENAI_PROVIDER}/{MODEL_NAME_GPT_4O_2024_11_20}": (
        "command_prompt_v2_gpt_4o_2024_11_20_template.jinja2"
    ),
    f"{AZURE_OPENAI_PROVIDER}/{MODEL_NAME_GPT_4O_2024_11_20}": (
        "command_prompt_v2_gpt_4o_2024_11_20_template.jinja2"
    ),
    f"{AWS_BEDROCK_PROVIDER}/anthropic.{MODEL_NAME_CLAUDE_3_5_SONNET_20240620}-v1:0": (
        "command_prompt_v2_claude_3_5_sonnet_20240620_template.jinja2"
    ),
    f"{ANTHROPIC_PROVIDER}/{MODEL_NAME_CLAUDE_3_5_SONNET_20240620}": (
        "command_prompt_v2_claude_3_5_sonnet_20240620_template.jinja2"
    ),
}

# When model is not configured, then we use the default prompt template
DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME = (
    "command_prompt_v2_gpt_4o_2024_11_20_template.jinja2"
)
# When the configured model is not found in the model prompt mapper, then we use the
# fallback prompt template
FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME = (
    "command_prompt_v2_gpt_4o_2024_11_20_template.jinja2"
)


class CommandParserValidatorSingleton:
    """Singleton class to validate the command parser.

    This class is used to validate the command parser. It keeps track of the number of
    consecutive turns where no commands are parsed by the command parser. If the
    number of consecutive turns exceeds a certain threshold, a warning is logged.
    The prompt can use a DSL syntax that can be incompatible with the command syntax
    used by the command parser. This class helps to detect such incompatibilities.
    """

    MAX_CONSECUTIVE_TURNS_NO_COMMAND_PREDICTED = 5
    _NO_COMMAND_PREDICTED_TURN_COUNTER = 0
    _command_parser_validated = False

    @classmethod
    def get_no_command_predicted_turn_counter(cls) -> int:
        return cls._NO_COMMAND_PREDICTED_TURN_COUNTER

    @classmethod
    def should_validate_command_parser(cls) -> bool:
        return not cls._command_parser_validated

    @classmethod
    def reset_command_parser_validation(cls) -> None:
        cls._NO_COMMAND_PREDICTED_TURN_COUNTER = 0
        cls._command_parser_validated = False

    @classmethod
    def validate_if_commands_are_parsed_from_llm_response(
        cls, commands: List[Command], llm_response: str
    ) -> None:
        if llm_response and not commands:
            cls._NO_COMMAND_PREDICTED_TURN_COUNTER += 1
        else:
            # Reset the counter if commands are generated, and mark
            # the command parser as validated.
            cls._NO_COMMAND_PREDICTED_TURN_COUNTER = 0
            cls._command_parser_validated = True
            return

        if (
            cls._NO_COMMAND_PREDICTED_TURN_COUNTER
            >= cls.MAX_CONSECUTIVE_TURNS_NO_COMMAND_PREDICTED
        ):
            structlogger.warning(
                "llm_command_generator.predict_commands.command_parser_not_working",
                event_info=(
                    f"No commands were generated by the command parser for the last "
                    f"{cls._NO_COMMAND_PREDICTED_TURN_COUNTER} times. Check if you "
                    "are running incompatible prompt and LLM command generator."
                ),
            )


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
class CompactLLMCommandGenerator(LLMBasedCommandGenerator):
    """A single step LLM-based command generator."""

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            config,
            model_storage,
            resource,
            prompt_template=prompt_template,
            **kwargs,
        )

        # Get the prompt template from the config or the default prompt template.
        self.prompt_template = self._resolve_component_prompt_template(
            self.config, prompt_template, log_context=LOG_COMPONENT_SOURCE_METHOD_INIT
        )

        # Set the command syntax version to v2
        CommandSyntaxManager.set_syntax_version(
            self.get_component_command_syntax_version()
        )

        self.trace_prompt_tokens = self.config.get("trace_prompt_tokens", False)
        self.repeat_command_enabled = self.is_repeat_command_enabled()

    ### Implementations of LLMBasedCommandGenerator parent
    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            PROMPT_TEMPLATE_CONFIG_KEY: None,
            USER_INPUT_CONFIG_KEY: None,
            LLM_CONFIG_KEY: None,
            FLOW_RETRIEVAL_KEY: FlowRetrieval.get_default_config(),
        }

    def persist(self) -> None:
        """Persist this component to disk for future loading."""
        self._persist_prompt_template()
        self._persist_config()
        if self.flow_retrieval is not None:
            self.flow_retrieval.persist()

    def _persist_prompt_template(self) -> None:
        """Persist prompt template for future loading."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.write_text_file(
                self.prompt_template, path / COMMAND_PROMPT_FILE_NAME
            )

    def _persist_config(self) -> None:
        """Persist config as a source of truth for resolved clients."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                path / LLM_BASED_COMMAND_GENERATOR_CONFIG_FILE, self.config
            )

    @classmethod
    def load(
        cls: Any,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "CompactLLMCommandGenerator":
        """Loads trained component (see parent class for full docstring)."""
        # Perform health check of the LLM API endpoint
        llm_config = resolve_model_client_config(config.get(LLM_CONFIG_KEY, {}))
        cls.perform_llm_health_check(
            llm_config,
            cls.get_default_llm_config(),
            "compact_llm_command_generator.load",
            cls.__name__,
        )

        # load prompt template from the model storage.
        prompt_template = cls.load_prompt_template_from_model_storage(
            model_storage, resource, COMMAND_PROMPT_FILE_NAME
        )

        # init base command generator
        command_generator = cls(config, model_storage, resource, prompt_template)
        # load flow retrieval if enabled
        if command_generator.enabled_flow_retrieval:
            command_generator.flow_retrieval = cls.load_flow_retrival(
                command_generator.config, model_storage, resource
            )

        return command_generator

    async def predict_commands(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
        **kwargs: Any,
    ) -> List[Command]:
        """Predict commands using the LLM.

        Args:
            message: The message from the user.
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.
            **kwargs: Keyword arguments for forward compatibility.

        Returns:
            The commands generated by the llm.
        """
        prior_commands = self._get_prior_commands(message)

        if tracker is None or flows.is_empty():
            # cannot do anything if there are no flows or no tracker
            return prior_commands

        if self._should_skip_llm_call(prior_commands, flows, tracker):
            return prior_commands

        try:
            commands = await self._predict_commands(message, flows, tracker)
        except ProviderClientAPIException:
            # if command predictions resulted in API exception
            # "predict" the ErrorCommand
            commands = [ErrorCommand()]
            structlogger.warning(
                "llm_command_generator.predict_commands.api_exception",
                event_info=(
                    "ProviderClientAPIException occurred while predicting commands."
                ),
                commands=commands,
            )

        if not commands and not prior_commands:
            # no commands are parsed or there's an invalid command
            structlogger.warning(
                "llm_command_generator.predict_commands",
                message="No commands were predicted as the LLM response could "
                "not be parsed or the LLM responded with an invalid command. "
                "Returning a CannotHandleCommand instead.",
            )
            commands = [CannotHandleCommand()]

        if tracker.has_coexistence_routing_slot:
            # if coexistence feature is used, set the routing slot
            commands += [SetSlotCommand(ROUTE_TO_CALM_SLOT, True)]

        log_llm(
            logger=structlogger,
            log_module=self.__class__.__name__,
            log_event="llm_command_generator.predict_commands.finished",
            commands=commands,
        )

        domain = kwargs.get("domain")
        commands = self._check_commands_against_slot_mappings(commands, tracker, domain)

        return self._check_commands_overlap(prior_commands, commands)

    async def _predict_commands(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
    ) -> List[Command]:
        """Predict commands using the LLM.

        Args:
            message: The message from the user.
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.

        Returns:
            The commands generated by the llm.

        Raises:
            ProviderClientAPIException: If API calls raised an error.
        """
        # retrieve flows
        filtered_flows = await self.filter_flows(message, flows, tracker)

        flow_prompt = self.render_template(message, tracker, filtered_flows, flows)
        log_llm(
            logger=structlogger,
            log_module=self.__class__.__name__,
            log_event="llm_command_generator.predict_commands.prompt_rendered",
            prompt=flow_prompt,
        )

        response = await self.invoke_llm(flow_prompt)
        llm_response = LLMResponse.ensure_llm_response(response)
        # The check for 'None' maintains compatibility with older versions
        # of LLMCommandGenerator. In previous implementations, 'invoke_llm'
        # might return 'None' to indicate a failure to generate actions.
        if llm_response is None or not llm_response.choices:
            structlogger.warning(
                "llm_command_generator.predict_commands.no_actions_generated",
                event_info=(
                    "No actions were generated by the LLM. Returning an ErrorCommand."
                ),
            )
            return [ErrorCommand()]

        action_list = llm_response.choices[0]

        log_llm(
            logger=structlogger,
            log_module=self.__class__.__name__,
            log_event="llm_command_generator.predict_commands.actions_generated",
            action_list=action_list,
        )

        commands = self.parse_commands(action_list, tracker, flows)

        if CommandParserValidatorSingleton.should_validate_command_parser():
            CommandParserValidatorSingleton.validate_if_commands_are_parsed_from_llm_response(
                commands, action_list
            )

        self._update_message_parse_data_for_fine_tuning(message, commands, flow_prompt)
        add_commands_to_message_parse_data(message, self.__class__.__name__, commands)
        add_prompt_to_message_parse_data(
            message=message,
            component_name=self.__class__.__name__,
            prompt_name="command_generator_prompt",
            user_prompt=flow_prompt,
            llm_response=llm_response,
        )

        return commands

    @staticmethod
    def _update_message_parse_data_for_fine_tuning(
        message: Message, commands: List[Command], prompt: str
    ) -> None:
        from rasa.llm_fine_tuning.annotation_module import preparing_fine_tuning_data

        if preparing_fine_tuning_data:
            # Add commands and prompt to the message object in order to create
            # prompt -> commands pairs for fine-tuning
            message.set(
                LLM_COMMANDS,
                [command.as_dict() for command in commands],
                add_to_output=True,
            )
            message.set(LLM_PROMPT, prompt, add_to_output=True)

    @classmethod
    def parse_commands(
        cls, actions: Optional[str], tracker: DialogueStateTracker, flows: FlowsList
    ) -> List[Command]:
        """Parse the actions returned by the llm into intent and entities.

        Args:
            actions: The actions returned by the llm.
            tracker: The tracker containing the current state of the conversation.
            flows: the list of flows

        Returns:
            The parsed commands.
        """
        commands = parse_commands_using_command_parsers(actions, flows)
        if not commands:
            structlogger.warning(
                f"{cls.__name__}.parse_commands",
                message="No commands were parsed from the LLM actions.",
                actions=actions,
            )

        return commands

    ### Helper methods
    def render_template(
        self,
        message: Message,
        tracker: DialogueStateTracker,
        startable_flows: FlowsList,
        all_flows: FlowsList,
    ) -> str:
        """Render the jinja template to create the prompt for the LLM.

        Args:
            message: The current message from the user.
            tracker: The tracker containing the current state of the conversation.
            startable_flows: The flows startable at this point in time by the user.
            all_flows: all flows present in the assistant

        Returns:
            The rendered prompt template.
        """
        # need to make this distinction here because current step of the
        # top_calling_frame would be the call step, but we need the collect step from
        # the called frame. If no call is active calling and called frame are the same.
        top_calling_frame = top_flow_frame(tracker.stack)
        top_called_frame = top_flow_frame(tracker.stack, ignore_call_frames=False)

        top_flow = top_calling_frame.flow(all_flows) if top_calling_frame else None
        current_step = top_called_frame.step(all_flows) if top_called_frame else None

        flow_slots = self.prepare_current_flow_slots_for_template(
            top_flow, current_step, tracker
        )
        current_slot, current_slot_description = self.prepare_current_slot_for_template(
            current_step
        )
        current_slot_type = None
        current_slot_allowed_values = None
        if current_slot:
            current_slot_type = (
                slot.type_name
                if (slot := tracker.slots.get(current_slot)) is not None
                else None
            )
            current_slot_allowed_values = allowed_values_for_slot(
                tracker.slots.get(current_slot)
            )
        current_conversation = tracker_as_readable_transcript(tracker)
        latest_user_message = sanitize_message_for_prompt(message.get(TEXT))
        current_conversation += f"\nUSER: {latest_user_message}"

        inputs = {
            "available_flows": self.prepare_flows_for_template(
                startable_flows, tracker
            ),
            "current_conversation": current_conversation,
            "flow_slots": flow_slots,
            "current_flow": top_flow.id if top_flow is not None else None,
            "current_slot": current_slot,
            "current_slot_description": current_slot_description,
            "current_slot_type": current_slot_type,
            "current_slot_allowed_values": current_slot_allowed_values,
            "user_message": latest_user_message,
            "is_repeat_command_enabled": self.repeat_command_enabled,
        }

        return self.compile_template(self.prompt_template).render(**inputs)

    def is_repeat_command_enabled(self) -> bool:
        """Check for feature flag"""
        RASA_PRO_BETA_REPEAT_COMMAND_ENV_VAR_NAME = "RASA_PRO_BETA_REPEAT_COMMAND"
        try:
            ensure_beta_feature_is_enabled(
                "Repeat Command",
                env_flag=RASA_PRO_BETA_REPEAT_COMMAND_ENV_VAR_NAME,
            )
        except BetaNotEnabledException:
            return False

        return True

    @classmethod
    def fingerprint_addon(cls: Any, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint for the graph."""
        # Get the default prompt template based on the model name
        llm_config = resolve_model_client_config(
            config.get(LLM_CONFIG_KEY), CompactLLMCommandGenerator.__name__
        )
        embedding_config = resolve_model_client_config(
            config.get(FLOW_RETRIEVAL_KEY, {}).get(EMBEDDINGS_CONFIG_KEY),
            FlowRetrieval.__name__,
        )

        # Create a copy of the config to avoid modifying the original config
        # and update the llm config with the resolved llm config.
        _config_copy = copy.deepcopy(config)
        _config_copy[LLM_CONFIG_KEY] = llm_config
        prompt_template = cls._resolve_component_prompt_template(
            _config_copy, log_context=LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON
        )

        return deep_container_fingerprint(
            [prompt_template, llm_config, embedding_config]
        )

    @staticmethod
    def get_default_llm_config() -> Dict[str, Any]:
        """Get the default LLM config for the command generator."""
        return DEFAULT_LLM_CONFIG

    @staticmethod
    def get_component_command_syntax_version() -> CommandSyntaxVersion:
        return CommandSyntaxVersion.v2

    @staticmethod
    def _resolve_component_prompt_template(
        config: Dict[str, Any],
        prompt_template: Optional[str] = None,
        log_context: Optional[Literal["init", "fingerprint_addon"]] = None,
    ) -> Optional[str]:
        """Get the prompt template from the config or the default prompt template."""
        # Get the default prompt template based on the model name.
        default_command_prompt_template = get_default_prompt_template_based_on_model(
            llm_config=config.get(LLM_CONFIG_KEY, {}) or {},
            model_prompt_mapping=MODEL_PROMPT_MAPPER,
            default_prompt_path=DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME,
            fallback_prompt_path=FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME,
            log_source_component=CompactLLMCommandGenerator.__name__,
            log_source_method=log_context,
        )

        # Return the prompt template either from the config or the default prompt.
        return prompt_template or get_prompt_template(
            config.get(PROMPT_TEMPLATE_CONFIG_KEY),
            default_command_prompt_template,
            log_source_component=CompactLLMCommandGenerator.__name__,
            log_source_method=log_context,
        )



from __future__ import annotations

import dataclasses
import logging
import typing
from typing import Any, Dict, List, Optional, Text, Tuple

from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.nlu.constants import DENSE_FEATURIZABLE_ATTRIBUTES, SPACY_DOCS
from rasa.nlu.model import InvalidModelError
from rasa.shared.constants import DOCS_URL_COMPONENTS
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from spacy.language import Language
    from spacy.tokens import Doc


@dataclasses.dataclass
class SpacyModel:
    """Wraps `SpacyNLP` output to make it fingerprintable."""

    model: Language
    model_name: Text

    def fingerprint(self) -> Text:
        """Fingerprints the model name.

        Use a static fingerprint as we assume this only changes if the model name
        changes and want to avoid investigating the model in greater detail for now.

        Returns:
            Fingerprint for model.
        """
        return str(self.model_name)


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.MODEL_LOADER,
        DefaultV1Recipe.ComponentType.MESSAGE_FEATURIZER,
    ],
    is_trainable=False,
    model_from="SpacyNLP",
)
class SpacyNLP(GraphComponent):
    """Component which provides the common loaded SpaCy model to others.

    This is used to avoid loading the SpaCy model multiple times. Instead the Spacy
    model is only loaded once and then shared by depending components.
    """

    def __init__(self, model: SpacyModel, config: Dict[Text, Any]) -> None:
        """Initializes a `SpacyNLP`."""
        self._model = model
        self._config = config

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Default config."""
        return {
            # when retrieving word vectors, this will decide if the casing
            # of the word is relevant. E.g. `hello` and `Hello` will
            # retrieve the same vector, if set to `False`. For some
            # applications and models it makes sense to differentiate
            # between these two words, therefore setting this to `True`.
            "case_sensitive": False
        }

    @staticmethod
    def load_model(spacy_model_name: Text) -> SpacyModel:
        """Try loading the model, catching the OSError if missing."""
        import spacy

        if not spacy_model_name:
            raise InvalidModelError(
                f"Missing model configuration for `SpacyNLP` in `config.yml`.\n"
                f"You must pass a model to the `SpacyNLP` component explicitly.\n"
                f"For example:\n"
                f"- name: SpacyNLP\n"
                f"  model: en_core_web_md\n"
                f"More information can be found on {DOCS_URL_COMPONENTS}#spacynlp"
            )

        try:
            language = spacy.load(spacy_model_name, disable=["parser"])
            spacy_runtime_version = spacy.about.__version__
            spacy_model_info = spacy.info(spacy_model_name)
            spacy_model_version_req = (
                spacy_model_info.get("spacy_version")
                if isinstance(spacy_model_info, dict)
                else ""
            )
            if not spacy.util.is_compatible_version(
                spacy_runtime_version, spacy_model_version_req
            ):
                raise InvalidModelError(
                    f"The specified model - {spacy_model_name} requires a spaCy "
                    f"runtime version {spacy_model_version_req} and is not compatible "
                    f"with the current spaCy runtime version {spacy_runtime_version}"
                )
            return SpacyModel(model=language, model_name=spacy_model_name)
        except OSError:
            raise InvalidModelError(
                f"Please confirm that {spacy_model_name} is an available spaCy model. "
                f"You need to download one upfront. For example:\n"
                f"python -m spacy download en_core_web_md\n"
                f"More information can be found on {DOCS_URL_COMPONENTS}#spacynlp"
            )

    @staticmethod
    def required_packages() -> List[Text]:
        """Lists required dependencies (see parent class for full docstring)."""
        return ["spacy"]

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> SpacyNLP:
        """Creates component (see parent class for full docstring)."""
        spacy_model_name = config.get("model")

        logger.info(f"Trying to load SpaCy model with name '{spacy_model_name}'.")

        model = cls.load_model(spacy_model_name)

        cls.ensure_proper_language_model(model.model)
        return cls(model, {**cls.get_default_config(), **config})

    @staticmethod
    def ensure_proper_language_model(nlp: Optional[Language]) -> None:
        """Checks if the SpaCy language model is properly loaded.

        Raises an exception if the model is invalid.
        """
        if nlp is None:
            raise Exception(
                "Failed to load SpaCy language model. "
                "Loading the model returned 'None'."
            )
        if nlp.path is None:
            # Spacy sets the path to `None` if
            # it did not load the model from disk.
            # In this case `nlp` is an unusable stub.
            raise Exception(
                f"Failed to load SpaCy language model for "
                f"lang '{nlp.lang}'. Make sure you have downloaded the "
                f"correct model (https://spacy.io/docs/usage/)."
                ""
            )

    def provide(self) -> SpacyModel:
        """Provides the loaded SpaCy model."""
        return self._model

    def _doc_for_text(self, model: Language, text: Text) -> Doc:
        """Makes a SpaCy doc object from a string of text."""
        return model(self._preprocess_text(text))

    def _preprocess_text(self, text: Optional[Text]) -> Text:
        """Processes the text before it is handled by SpaCy."""
        if text is None:
            # converted to empty string so that it can still be passed to spacy.
            # Another option could be to neglect tokenization of the attribute of
            # this example, but since we are processing in batch mode, it would
            # get complex to collect all processed and neglected examples.
            text = ""
        if self._config.get("case_sensitive"):
            return text
        else:
            return text.lower()

    def _get_text(self, example: Dict[Text, Any], attribute: Text) -> Text:
        return self._preprocess_text(example.get(attribute))

    @staticmethod
    def _merge_content_lists(
        indexed_training_samples: List[Tuple[int, Text]],
        doc_lists: List[Tuple[int, Doc]],
    ) -> List[Tuple[int, Doc]]:
        """Merge lists with processed Docs back into their original order."""
        dct = dict(indexed_training_samples)
        dct.update(doc_lists)
        return sorted(dct.items())

    @staticmethod
    def _filter_training_samples_by_content(
        indexed_training_samples: List[Tuple[int, Text]],
    ) -> Tuple[List[Tuple[int, Text]], List[Tuple[int, Text]]]:
        """Separates empty training samples from content bearing ones."""
        docs_to_pipe = list(
            filter(
                lambda training_sample: training_sample[1] != "",
                indexed_training_samples,
            )
        )
        empty_docs = list(
            filter(
                lambda training_sample: training_sample[1] == "",
                indexed_training_samples,
            )
        )
        return docs_to_pipe, empty_docs

    @staticmethod
    def _process_content_bearing_samples(
        model: Language, samples_to_pipe: List[Tuple[int, Text]]
    ) -> List[Tuple[int, Doc]]:
        """Sends content bearing training samples to SpaCy's pipe."""
        docs = [
            (to_pipe_sample[0], doc)
            for to_pipe_sample, doc in zip(
                samples_to_pipe,
                [
                    doc
                    for doc in model.pipe(
                        [txt for _, txt in samples_to_pipe], batch_size=50
                    )
                ],
            )
        ]
        return docs

    @staticmethod
    def _process_non_content_bearing_samples(
        model: Language, empty_samples: List[Tuple[int, Text]]
    ) -> List[Tuple[int, Doc]]:
        """Creates empty Doc-objects from zero-lengthed training samples strings."""
        from spacy.tokens import Doc

        n_docs = [
            (empty_sample[0], doc)
            for empty_sample, doc in zip(
                empty_samples, [Doc(model.vocab) for doc in empty_samples]
            )
        ]
        return n_docs

    def _docs_for_training_data(
        self, model: Language, training_data: TrainingData
    ) -> Dict[Text, List[Any]]:
        attribute_docs = {}
        for attribute in DENSE_FEATURIZABLE_ATTRIBUTES:
            texts = [
                self._get_text(e, attribute) for e in training_data.training_examples
            ]
            # Index and freeze indices of the training samples for preserving the order
            # after processing the data.
            indexed_training_samples = [(idx, text) for idx, text in enumerate(texts)]

            samples_to_pipe, empty_samples = self._filter_training_samples_by_content(
                indexed_training_samples
            )

            content_bearing_docs = self._process_content_bearing_samples(
                model, samples_to_pipe
            )

            non_content_bearing_docs = self._process_non_content_bearing_samples(
                model, empty_samples
            )

            attribute_document_list = self._merge_content_lists(
                indexed_training_samples,
                content_bearing_docs + non_content_bearing_docs,
            )

            # Since we only need the training samples strings,
            # we create a list to get them out of the tuple.
            attribute_docs[attribute] = [doc for _, doc in attribute_document_list]
        return attribute_docs

    def process_training_data(
        self, training_data: TrainingData, model: SpacyModel
    ) -> TrainingData:
        """Adds SpaCy tokens and features to training data messages."""
        attribute_docs = self._docs_for_training_data(model.model, training_data)

        for attribute in DENSE_FEATURIZABLE_ATTRIBUTES:
            for idx, example in enumerate(training_data.training_examples):
                example_attribute_doc = attribute_docs[attribute][idx]
                if len(example_attribute_doc):
                    # If length is 0, that means the initial text feature
                    # was None and was replaced by ''
                    # in preprocess method
                    example.set(SPACY_DOCS[attribute], example_attribute_doc)

        return training_data

    def process(self, messages: List[Message], model: SpacyModel) -> List[Message]:
        """Adds SpaCy tokens and features to messages."""
        for message in messages:
            for attribute in DENSE_FEATURIZABLE_ATTRIBUTES:
                if message.get(attribute):
                    message.set(
                        SPACY_DOCS[attribute],
                        self._doc_for_text(model.model, message.get(attribute)),
                    )

        return messages



from __future__ import annotations

import copy
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Text, Union

import structlog
from pydantic import BaseModel
from pypred import Predicate

import rasa.shared.utils.io
from rasa.engine.language import Language
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.flows.constants import (
    KEY_ALWAYS_INCLUDE_IN_PROMPT,
    KEY_DESCRIPTION,
    KEY_FILE_PATH,
    KEY_ID,
    KEY_IF,
    KEY_NAME,
    KEY_NLU_TRIGGER,
    KEY_PERSISTED_SLOTS,
    KEY_RUN_PATTERN_COMPLETED,
    KEY_STEPS,
    KEY_TRANSLATION,
)
from rasa.shared.core.flows.flow_path import FlowPath, FlowPathsList, PathNode
from rasa.shared.core.flows.flow_step import FlowStep
from rasa.shared.core.flows.flow_step_links import (
    ElseFlowStepLink,
    FlowStepLink,
    IfFlowStepLink,
    StaticFlowStepLink,
)
from rasa.shared.core.flows.flow_step_sequence import FlowStepSequence
from rasa.shared.core.flows.nlu_trigger import NLUTriggers
from rasa.shared.core.flows.steps import (
    ActionFlowStep,
    CallFlowStep,
    CollectInformationFlowStep,
    EndFlowStep,
    StartFlowStep,
)
from rasa.shared.core.flows.steps.constants import (
    CONTINUE_STEP_PREFIX,
    END_STEP,
    START_STEP,
)
from rasa.shared.core.flows.steps.continuation import ContinueFlowStep
from rasa.shared.core.slots import Slot

structlogger = structlog.get_logger()


class FlowLanguageTranslation(BaseModel):
    """Represents the translation of the flow properties in a specific language."""

    name: str
    """The human-readable name of the flow."""

    class Config:
        extra = "ignore"


@dataclass
class Flow:
    """Represents the configuration of a flow."""

    id: Text
    """The id of the flow."""
    custom_name: Optional[Text] = None
    """The human-readable name of the flow."""
    description: Optional[Text] = None
    """The description of the flow."""
    translation: Dict[Text, FlowLanguageTranslation] = field(default_factory=dict)
    """The translation of the flow properties in different languages."""
    guard_condition: Optional[Text] = None
    """The condition that needs to be fulfilled for the flow to be startable."""
    step_sequence: FlowStepSequence = field(default_factory=FlowStepSequence.empty)
    """The steps of the flow."""
    nlu_triggers: Optional[NLUTriggers] = None
    """The list of intents, e.g. nlu triggers, that start the flow."""
    always_include_in_prompt: Optional[bool] = None
    """
    A flag that checks whether the flow should always be included in the prompt or not.
    """
    file_path: Optional[str] = None
    """The path to the file where the flow is stored."""
    persisted_slots: List[str] = field(default_factory=list)
    """The list of slots that should be persisted after the flow ends."""
    run_pattern_completed: bool = True
    """Whether the pattern_completed flow should be run after the flow ends."""

    @staticmethod
    def from_json(
        flow_id: Text,
        data: Dict[Text, Any],
        file_path: Optional[Union[str, Path]] = None,
    ) -> Flow:
        """Create a Flow object from serialized data.

        Args:
            flow_id: id of the flow
            data: data for a Flow object in a serialized format.
            file_path: the file path of the flow

        Returns:
            A Flow object.
        """
        from rasa.shared.core.flows.utils import extract_translations

        step_sequence = FlowStepSequence.from_json(flow_id, data.get("steps"))
        nlu_triggers = NLUTriggers.from_json(data.get("nlu_trigger"))

        if file_path and isinstance(file_path, Path):
            file_path = str(file_path)

        return Flow(
            id=flow_id,
            custom_name=data.get(KEY_NAME),
            description=data.get(KEY_DESCRIPTION),
            always_include_in_prompt=data.get(KEY_ALWAYS_INCLUDE_IN_PROMPT),
            # str or bool are permitted in the flow schema, but internally we want a str
            guard_condition=str(data[KEY_IF]) if KEY_IF in data else None,
            step_sequence=Flow.resolve_default_ids(step_sequence),
            nlu_triggers=nlu_triggers,
            # If we are reading the flows in after training the file_path is part of
            # data. When the model is trained, take the provided file_path.
            file_path=data.get(KEY_FILE_PATH) if KEY_FILE_PATH in data else file_path,
            persisted_slots=data.get(KEY_PERSISTED_SLOTS, []),
            run_pattern_completed=data.get(KEY_RUN_PATTERN_COMPLETED, True),
            translation=extract_translations(
                translation_data=data.get(KEY_TRANSLATION, {})
            ),
        )

    def get_full_name(self) -> str:
        if self.file_path:
            return f"{self.file_path}::{self.name}"
        return self.name

    @staticmethod
    def create_default_name(flow_id: str) -> str:
        """Create a default flow name for when it is missing."""
        return flow_id.replace("_", " ").replace("-", " ")

    @staticmethod
    def resolve_default_ids(step_sequence: FlowStepSequence) -> FlowStepSequence:
        """Resolves the default ids of all steps in the sequence.

        If a step does not have an id, a default id is assigned to it based
        on the type of the step and its position in the flow.

        Similarly, if a step doesn't have an explicit next assigned we resolve
        the default next step id.

        Args:
            step_sequence: The step sequence to resolve the default ids for.

        Returns:
            The step sequence with the default ids resolved.
        """
        # assign an index to all steps
        for idx, step in enumerate(step_sequence.steps):
            step.idx = idx

        def resolve_default_next(steps: List[FlowStep], is_root_sequence: bool) -> None:
            for i, step in enumerate(steps):
                if step.next.no_link_available() and step.does_allow_for_next_step():
                    if i == len(steps) - 1:
                        if is_root_sequence:
                            # if this is the root sequence, we need to add an end step
                            # to the end of the sequence. other sequences, e.g.
                            # in branches need to explicitly add a next step.
                            step.next.links.append(StaticFlowStepLink(END_STEP))
                    else:
                        step.next.links.append(StaticFlowStepLink(steps[i + 1].id))
                for link in step.next.links:
                    if sub_steps := link.child_steps():
                        resolve_default_next(sub_steps, is_root_sequence=False)

        resolve_default_next(step_sequence.child_steps, is_root_sequence=True)
        return step_sequence

    def as_json(self) -> Dict[Text, Any]:
        """Serialize the Flow object.

        Returns:
            The Flow object as serialized data.
        """
        data: Dict[Text, Any] = {
            KEY_ID: self.id,
            KEY_STEPS: self.step_sequence.as_json(),
        }
        if self.custom_name is not None:
            data[KEY_NAME] = self.custom_name
        if self.description is not None:
            data[KEY_DESCRIPTION] = self.description
        if self.guard_condition is not None:
            data[KEY_IF] = self.guard_condition
        if self.always_include_in_prompt is not None:
            data[KEY_ALWAYS_INCLUDE_IN_PROMPT] = self.always_include_in_prompt
        if self.nlu_triggers:
            data[KEY_NLU_TRIGGER] = self.nlu_triggers.as_json()
        if self.file_path:
            data[KEY_FILE_PATH] = self.file_path
        if self.persisted_slots:
            data[KEY_PERSISTED_SLOTS] = self.persisted_slots
        if self.run_pattern_completed is not None:
            data["run_pattern_completed"] = self.run_pattern_completed
        if self.translation:
            data[KEY_TRANSLATION] = {
                language_code: translation.dict()
                for language_code, translation in self.translation.items()
            }

        return data

    def localized_name(self, language: Optional[Language] = None) -> Optional[Text]:
        """Returns the language specific flow name or None.

        Args:
            language: Preferred language code.

        Returns:
            Flow name in the specified language or None.
        """
        language_code = language.code if language else None
        translation = self.translation.get(language_code)
        return translation.name if translation else None

    def readable_name(self, language: Optional[Language] = None) -> str:
        """
        Returns the flow's name in the specified language if available; otherwise
        falls back to the flow's name, and finally the flow's ID.

        Args:
            language: Preferred language code.

        Returns:
            string: the localized name, the default name, or the flow's ID.
        """
        return self.localized_name(language) or self.name or self.id

    def step_by_id(self, step_id: Optional[Text]) -> Optional[FlowStep]:
        """Returns the step with the given id."""
        if not step_id:
            return None

        if step_id == START_STEP:
            return StartFlowStep(self.id, self.first_step_in_flow().id)

        if step_id == END_STEP:
            return EndFlowStep(self.id)

        if step_id.startswith(CONTINUE_STEP_PREFIX):
            return ContinueFlowStep(self.id, step_id[len(CONTINUE_STEP_PREFIX) :])

        for step in self.steps_with_calls_resolved:
            if step.id == step_id:
                return step

        return None

    def first_step_in_flow(self) -> FlowStep:
        """Returns the start step of this flow."""
        if not (steps := self.steps):
            raise RuntimeError(
                f"Flow {self.id} is empty despite validation that this cannot happen."
            )
        return steps[0]

    def get_trigger_intents(self) -> Set[str]:
        """Returns the trigger intents of the flow."""
        results: Set[str] = set()

        if not self.nlu_triggers:
            return results

        for condition in self.nlu_triggers.trigger_conditions:
            results.add(condition.intent)

        return results

    @property
    def is_rasa_default_flow(self) -> bool:
        """Test whether the flow is a rasa default flow."""
        return self.id.startswith(RASA_DEFAULT_FLOW_PATTERN_PREFIX)

    def get_collect_steps(self) -> List[CollectInformationFlowStep]:
        """Return all CollectInformationFlowSteps in the flow."""
        collect_steps = []
        for step in self.steps_with_calls_resolved:
            if isinstance(step, CollectInformationFlowStep):
                collect_steps.append(step)
        return collect_steps

    @property
    def steps_with_calls_resolved(self) -> List[FlowStep]:
        """Return the steps of the flow including steps of called flows."""
        return self.step_sequence.steps_with_calls_resolved

    @property
    def steps(self) -> List[FlowStep]:
        """Return the steps of the flow without steps of called flows."""
        return self.step_sequence.steps

    @cached_property
    def fingerprint(self) -> str:
        """Create a fingerprint identifying this step sequence."""
        return rasa.shared.utils.io.deep_container_fingerprint(self.as_json())

    @property
    def utterances(self) -> Set[str]:
        """Retrieve all utterances of this flow."""
        return set().union(
            *[step.utterances for step in self.step_sequence.steps_with_calls_resolved]
        )

    @property
    def custom_actions(self) -> Set[str]:
        """Retrieve all custom actions of this flow."""
        return {
            step.custom_action
            for step in self.step_sequence.steps_with_calls_resolved
            if isinstance(step, ActionFlowStep) and step.custom_action is not None
        }

    @property
    def name(self) -> str:
        """Create a default name if none is present."""
        return self.custom_name or Flow.create_default_name(self.id)

    def is_startable(
        self,
        context: Optional[Dict[Text, Any]] = None,
        slots: Optional[Dict[Text, Slot]] = None,
    ) -> bool:
        """Return whether the start condition is satisfied.

        Args:
            context: The context data to evaluate the starting conditions against.
            slots: The slots to evaluate the starting conditions against.

        Returns:
            Whether the start condition is satisfied.
        """
        context = context or {}
        slots = slots or {}
        simplified_slots = {slot.name: slot.value for slot in slots.values()}

        # If no starting condition exists, the flow is always startable.
        if not self.guard_condition:
            return True

        # if a flow guard condition exists and the flow was started via a link,
        # e.g. is currently active, the flow is startable
        if context.get("flow_id") == self.id:
            return True

        try:
            predicate = Predicate(self.guard_condition)
            is_startable = predicate.evaluate(
                {"context": context, "slots": simplified_slots}
            )
            structlogger.debug(
                "command_generator.validate_flow_starting_conditions.result",
                predicate=predicate.description(),
                is_startable=is_startable,
                flow_id=self.id,
            )
            return is_startable
        # if there is any kind of exception when evaluating the predicate, the flow
        # is not startable
        except (TypeError, Exception) as e:
            structlogger.error(
                "command_generator.validate_flow_starting_conditions.error",
                predicate=self.guard_condition,
                context=context,
                slots=slots,
                error=str(e),
            )
            return False

    def has_action_step(self, action: Text) -> bool:
        """Check whether the flow has an action step with the given action."""
        for step in self.steps:
            if isinstance(step, ActionFlowStep) and step.action == action:
                return True
        return False

    def is_startable_only_via_link(self) -> bool:
        """Determines if the flow can be initiated exclusively through a link.

        This condition is met when a guard condition exists and is
        consistently evaluated to `False` (e.g. `if: False`).

        Returns:
            A boolean indicating if the flow initiation is link-based only.
        """
        if (
            self.guard_condition is None
            or self._contains_variables_in_guard_condition()
        ):
            return False

        try:
            predicate = Predicate(self.guard_condition)
            is_startable_via_link = not predicate.evaluate({})
            structlogger.debug(
                "flow.is_startable_only_via_link.result",
                predicate=self.guard_condition,
                is_startable_via_link=is_startable_via_link,
                flow_id=self.id,
            )
            return is_startable_via_link
        # if there is any kind of exception when evaluating the predicate, the flow
        # is not startable by link or by any other means.
        except (TypeError, Exception) as e:
            structlogger.error(
                "flow.is_startable_only_via_link.error",
                predicate=self.guard_condition,
                error=str(e),
                flow_id=self.id,
            )
            return False

    def _contains_variables_in_guard_condition(self) -> bool:
        """Determines if the guard condition contains dynamic literals.

        I.e. literals that cannot be statically resolved, indicating a variable.

        Returns:
            True if dynamic literals are present, False otherwise.
        """
        from pypred import ast
        from pypred.tiler import SimplePattern, tile

        if not self.guard_condition:
            return False

        predicate = Predicate(self.guard_condition)

        # find all literals in the AST tree
        literals = []
        tile(
            ast=predicate.ast,
            patterns=[SimplePattern("types:Literal")],
            func=lambda _, literal: literals.append(literal),
        )

        # check if there is a literal that cannot be statically resolved (variable)
        for literal in literals:
            if type(predicate.static_resolve(literal.value)) == ast.Undefined:
                return True

        return False

    def extract_all_paths(self) -> FlowPathsList:
        """Extracts all possible flow paths.

        Extracts all possible flow paths from a given flow structure by
        recursively exploring each step.
        This function initializes an empty list to collect paths, an empty path list,
        and a set of visited step IDs to prevent revisiting steps.
        It calls `go_over_steps` to recursively explore and fill the paths list.
        """
        all_paths = FlowPathsList(self.id, paths=[])
        start_step: FlowStep = self.first_step_in_flow()
        current_path: FlowPath = FlowPath(flow=self.id, nodes=[])
        visited_step_ids: Set[str] = set()

        self._go_over_steps(start_step, current_path, all_paths, visited_step_ids)

        structlogger.debug(
            "shared.core.flows.flow.extract_all_paths",
            comment="Extraction complete",
            number_of_paths=len(all_paths.paths),
            flow_name=self.name,
        )
        return all_paths

    def _go_over_steps(
        self,
        current_step: FlowStep,
        current_path: FlowPath,
        all_paths: FlowPathsList,
        visited_step_ids: Set[str],
    ) -> None:
        """Processes the flow steps recursively.

        Args:
            current_step: The current step being processed.
            current_path: The current path being constructed.
            all_paths: The list where completed paths are added.
            visited_step_ids: A set of steps that have been visited to avoid cycles.

        Returns:
            None: This function modifies all_paths in place by appending new paths
            as they are found.
        """
        # Check if the step is relevant for testable_paths extraction.
        # We only create new path nodes for ActionFlowStep, CallFlowStep and
        # CollectInformationFlowStep because these are externally visible
        # changes in the assistant's behaviour (trackable in the e2e tests).
        # For other flow steps, we only follow their links.
        # We decided to ignore calls to other flows in our coverage analysis.
        should_add_node = isinstance(
            current_step, (CollectInformationFlowStep, ActionFlowStep, CallFlowStep)
        )
        if should_add_node:
            # Add current step to the current path that is being constructed.
            current_path.nodes.append(
                PathNode(
                    flow=current_path.flow,
                    step_id=current_step.id,
                    lines=current_step.metadata["line_numbers"],
                )
            )

        if current_step.id in visited_step_ids or self.is_end_of_path(current_step):
            # Found a cycle, or reached an end step, do not proceed further.
            all_paths.paths.append(copy.deepcopy(current_path))
            # Remove the last node from the path if it was added.
            if should_add_node:
                current_path.nodes.pop()
            return

        # Mark current step as visited in this path.
        visited_step_ids.add(current_step.id)

        # Iterate over all links of the current step.
        for link in current_step.next.links:
            self._handle_link(
                current_path,
                all_paths,
                visited_step_ids,
                link,
            )

        # Backtrack the current step and remove it from the path.
        visited_step_ids.remove(current_step.id)

        # Remove the last node from the path if it was added.
        if should_add_node:
            current_path.nodes.pop()

    def _handle_link(
        self,
        current_path: FlowPath,
        all_paths: FlowPathsList,
        visited_step_ids: Set[str],
        link: FlowStepLink,
    ) -> None:
        """Handles the next step in a flow.

        Args:
            current_path: The current path being constructed.
            all_paths: The list where completed paths are added.
            visited_step_ids: A set of steps that have been visited to avoid cycles.
            link: The link to be followed.

        Returns:
            None: This function modifies all_paths in place by appending new paths
            as they are found.
        """
        # StaticFlowStepLink is a direct link to the next step.
        if isinstance(link, StaticFlowStepLink):
            # Find the step by its id and continue the path.
            if step := self._get_step_by_step_id(link.target_step_id):
                self._go_over_steps(
                    step,
                    current_path,
                    all_paths,
                    visited_step_ids,
                )
                return
        # IfFlowStepLink and ElseFlowStepLink are conditional links.
        elif isinstance(link, (IfFlowStepLink, ElseFlowStepLink)):
            if isinstance(link.target_reference, FlowStepSequence):
                # If the target is a FlowStepSequence, we need to go over all
                # child steps of the sequence.
                for child_step in link.target_reference.child_steps:
                    self._go_over_steps(
                        child_step,
                        current_path,
                        all_paths,
                        visited_step_ids,
                    )
                return
            else:
                # Find the step by its id and continue the path.
                if step := self._get_step_by_step_id(link.target_reference):
                    self._go_over_steps(
                        step,
                        current_path,
                        all_paths,
                        visited_step_ids,
                    )
                    return

    def is_end_of_path(self, step: FlowStep) -> bool:
        """Check if there is no path available from the current step."""
        if (
            len(step.next.links) == 1
            and isinstance(step.next.links[0], StaticFlowStepLink)
            and step.next.links[0].target == END_STEP
        ):
            return True
        return False

    def _get_step_by_step_id(
        self,
        step_id: Optional[str],
    ) -> Optional[FlowStep]:
        """Get a step by its id from a list of steps."""
        for step in self.steps:
            if step.id == step_id:
                return step
        return None



from pathlib import Path
from typing import Any, Dict, List, Optional, Text

import numpy as np
import structlog
from langchain.schema.embeddings import Embeddings
from langchain_community.vectorstores import FAISS

from rasa.shared.constants import REQUIRED_SLOTS_KEY
from rasa.shared.core.domain import KEY_RESPONSES_TEXT, Domain
from rasa.shared.utils.llm import AI

structlogger = structlog.get_logger()


def persist_faiss_vector_store(path: Path, vecor_store: Optional[FAISS]) -> None:
    """Persists the given vector store to the given path.

    Args:
        path: path to persist the vector store to
        vecor_store: vector store to persist
    """
    if vecor_store:
        vecor_store.save_local(str(path))


def load_faiss_vector_store(path: Path, embedder: Embeddings) -> Optional[FAISS]:
    """Loads the vector store from the given path.

    Args:
        path: path to load the vector store from
        embedder: embedder to use for the vector store

    Returns:
        The loaded vector store or None if the path does not exist.
    """
    if path.exists():
        return FAISS.load_local(
            str(path), embedder, allow_dangerous_deserialization=True
        )
    else:
        return None


def extract_participant_messages_from_transcript(
    transcript: Text, participant: Text = AI
) -> List[Text]:
    r"""Extracts repsonses from the transcript.

    Args:
        transcript: transcript text to process
        participant: participant prefix (AI or Human, AI by default)

    Example:
        >>> extract_participant_messages_from_transcript("USER: Hello\\nAI: Hi there!")
        ["Hi there!"]

    Returns:
        List of extracted responses.
    """
    prefix = participant + ": "
    return [
        line[len(prefix) :]
        for line in transcript.split("\n")
        if line.startswith(prefix)
    ]


def extract_ai_response_examples(
    responses: Dict[Text, List[Dict[Text, Any]]],
) -> List[str]:
    """Extracts the responses from the domain.

    Only responses that have a text field are extracted.

    Example:
        >>> responses = {
        ...     "utter_greet": [
        ...         {"text": "hello"},
        ...     ],
        ...     "utter_goodbye": [
        ...         {"text": "goodbye"},
        ...         {"text": ""},
        ...     ],
        ... }
        >>> extract_ai_response_examples(responses)
        ["hello", "goodbye"]

    Args:
        responses: the responses from the domain

    Returns:
    A list of responses.
    """
    example_ai_responses = []

    for variations in responses.values():
        if not variations:
            continue
        for variation in variations:
            if response_text := variation.get(KEY_RESPONSES_TEXT):
                example_ai_responses.append(response_text)

    return example_ai_responses


def response_for_template(
    template_name: str, responses: Dict[Text, List[Dict[Text, Any]]]
) -> Optional[str]:
    """Returns an interaction for the given template name from the responses.

    Args:
        template_name: the name of the template
        responses: the responses from the domain

    Returns:
    The response text or None if no response was found.
    """
    rsps = responses.get(template_name, [])
    usuable_responses = [r for r in rsps if r.get(KEY_RESPONSES_TEXT)]
    if usuable_responses:
        selected_response = np.random.choice(usuable_responses)  # type: ignore
        return selected_response.get(KEY_RESPONSES_TEXT)
    else:
        structlogger.warning(
            "response_template.not_found",
            template_name=template_name,
        )
        return None


def form_utterances_to_action(domain: Domain) -> Dict[str, str]:
    """Returns a dictionary of form utterances to their form name.

    Args:
        domain: the domain of the assistant


    Returns:
    A dictionary of form utterances to their form name.
    """
    form_slots = [
        (slot_name, form_name)
        for form_name, props in domain.forms.items()
        for slot_name in props.get(REQUIRED_SLOTS_KEY, [])
    ]

    return {f"utter_ask_{slot_name}": form_name for slot_name, form_name in form_slots}
